from typing import Any

import msgspec
from pydantic import BaseModel

__all__ = ["serialize", "to_string"]


def serialize(obj: BaseModel, indent: bool = True) -> str:
    return obj.model_dump_json(indent=indent)


def to_string(data: Any) -> str:
    return msgspec.json.format(msgspec.json.encode(data).decode("utf-8"))



import msgspec
import pytest
import respx

from lueur.make_id import make_id
from lueur.models import GCPProject
from lueur.platform.gcp.project import (
    list_organization_projects,
    list_all_projects,
)

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_list_organization_projects(
    gcp_org_id: str, gcp_project: str
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://cloudresourcemanager.googleapis.com/v3/projects",
        params={"parent": f"organizations/{gcp_org_id}"},
    ).respond(
        content=msgspec.json.encode({"projects": [{"name": gcp_project}]})
    )

    with patch_auth():
        assert await list_organization_projects(gcp_org_id) == [gcp_project]


@respx.mock
@pytest.mark.asyncio
async def test_list_all_projects(gcp_org_id: str, gcp_project: str) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://cloudresourcemanager.googleapis.com/v3/projects:search",
        params={"query": "state:ACTIVE"},
    ).respond(
        content=msgspec.json.encode(
            {
                "projects": [
                    {
                        "parent": f"organizations/{gcp_org_id}",
                        "name": gcp_project,
                        "displayName": "my org",
                    }
                ]
            }
        )
    )

    with patch_auth():
        assert await list_all_projects() == [
            GCPProject(
                organization=f"organizations/{gcp_org_id}",
                id=make_id(gcp_project),
                name="my org",
            )
        ]



import msgspec
import pytest
import respx

from lueur.platform.gcp.region import list_project_regions

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_list_project_regions(gcp_project: str) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://compute.googleapis.com/compute/v1/projects/{gcp_project}/regions",
        params={"fields": "items.name"},
    ).respond(content=msgspec.json.encode({"items": [{"name": "us-east1"}]}))

    with patch_auth():
        assert await list_project_regions(gcp_project) == ["us-east1"]



from typing import Any

import msgspec
import pytest
import respx

from lueur.platform.gcp.connector import explore_connector

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_explore_connector(
    gcp_project: str,
    gcp_region: str,
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://connectors.googleapis.com/v1/projects/{gcp_project}/locations/{gcp_region}/providers",
    ).respond(content=msgspec.json.encode({"providers": []}))

    respx.get(
        f"https://connectors.googleapis.com/v1/projects/{gcp_project}/locations/{gcp_region}/connectors",
    ).respond(content=msgspec.json.encode({"connectors": []}))

    respx.get(
        f"https://connectors.googleapis.com/v1/projects/{gcp_project}/locations/{gcp_region}/connections",
    ).respond(content=msgspec.json.encode({"connectors": []}))

    with patch_auth():
        resources = await explore_connector(gcp_project, gcp_region)



import re
from typing import Any, Pattern, cast
from urllib.parse import urlparse

from pydantic import BaseModel

from lueur.resource import iter_resource

__all__ = ["lb_config_from_url", "LoadBalancerConfig"]


class LoadBalancerConfig(BaseModel):
    project: str
    url: str
    service: str
    urlmap: str
    path_matcher: str
    path: str
    is_global: bool = False
    is_regional: bool = False


def lb_config_from_url(
    snapshot: dict[str, Any], url: str
) -> LoadBalancerConfig | None:
    """
    Extract the load balancer config for a given URL.
    """
    defaults = {
        "project": "",
        "url": url,
        "service": "",
        "urlmap": "",
        "path_matcher": "",
        "path": "",
        "is_global": False,
        "is_regional": False,
    }

    p = urlparse(url)
    host = p.hostname
    path = p.path

    for m in iter_resource(
        snapshot,
        f"$.resources[?@.struct.hostRules.*.hosts contains '{host}']",
    ):
        urlmap = cast(dict, m.value)
        defaults["is_global"] = (
            True if urlmap["meta"]["kind"].startswith("global-") else False
        )
        defaults["is_regional"] = not defaults["is_global"]
        defaults["project"] = urlmap["meta"]["project"]
        matches = extract_path_prefixes(urlmap)
        defaults.update(get_first_match(matches, path or "/"))

        return LoadBalancerConfig.model_validate(defaults)

    return None


###############################################################################
# Private functions
###############################################################################
def extract_path_prefixes(
    urlmap: dict[str, Any],
) -> dict[Pattern, dict[str, str]]:
    matches = {}

    for pm in urlmap["struct"]["pathMatchers"]:
        for rr in pm["routeRules"]:
            for mr in rr.get("matchRules", []):
                if "prefixMatch" in mr:
                    matches[re.compile(f"^{mr['prefixMatch']}")] = {
                        "urlmap": urlmap["meta"]["name"],
                        "service": get_higest_priority_service(
                            rr["routeAction"]["weightedBackendServices"]
                        ),
                        "path_matcher": pm["name"],
                        "path": mr["prefixMatch"],
                    }
                elif "fullPathMatch" in mr:
                    matches[re.compile(f"^{mr['fullPathMatch']}$")] = {
                        "urlmap": urlmap["meta"]["name"],
                        "service": get_higest_priority_service(
                            rr["routeAction"]["weightedBackendServices"]
                        ),
                        "path_matcher": pm["name"],
                        "path": mr["fullPathMatch"],
                    }
                elif "regexMatch" in mr:
                    matches[re.compile(mr["regexMatch"])] = {
                        "urlmap": urlmap["meta"]["name"],
                        "service": get_higest_priority_service(
                            rr["routeAction"]["weightedBackendServices"]
                        ),
                        "path_matcher": pm["name"],
                        "path": mr["regexMatch"],
                    }

    return matches


def get_higest_priority_service(services: list[dict[str, Any]]) -> str:
    return max(services, key=lambda x: x["weight"])["backendService"].rsplit(
        "/", 1
    )[-1]


def get_first_match(
    matches: dict[Pattern, dict[str, str]], path: str
) -> dict[str, str]:
    for k, v in matches.items():
        if k.match(path):
            return v

    return {}



# mypy: disable-error-code="func-returns-value"
import asyncio
import logging
from typing import Any

from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import Discovery, Meta
from lueur.platform.gcp.address import explore_addresses
from lueur.platform.gcp.cloudrun import expand_links as cloudrun_links
from lueur.platform.gcp.cloudrun import explore_cloudrun
from lueur.platform.gcp.compute import explore_compute
from lueur.platform.gcp.connector import explore_connector
from lueur.platform.gcp.dns import explore_dns
from lueur.platform.gcp.firewall import explore_firewalls
from lueur.platform.gcp.forwardingrule import explore_forwardingrules
from lueur.platform.gcp.gke import expand_links as gke_expand_links
from lueur.platform.gcp.gke import explore_gke
from lueur.platform.gcp.healthchecks import explore_health_checks
from lueur.platform.gcp.lb import expand_links as lb_expand_links
from lueur.platform.gcp.lb import explore_lb
from lueur.platform.gcp.memorystore import explore_memorystore
from lueur.platform.gcp.monitoring import expand_links as mon_expand_links
from lueur.platform.gcp.monitoring import explore_monitoring
from lueur.platform.gcp.securities import explore_securities
from lueur.platform.gcp.sql import expand_links as sql_expand_links
from lueur.platform.gcp.sql import explore_sql
from lueur.platform.gcp.storage import explore_storage
from lueur.platform.gcp.targetproxy import explore_target_proxies
from lueur.platform.gcp.vpc import expand_links as vpc_expand_links
from lueur.platform.gcp.vpc import explore_vpc

__all__ = ["explore", "expand_links"]
logger = logging.getLogger("lueur.lib")


async def explore(
    project: str, location: str | None = None, creds: Credentials | None = None
) -> Discovery:
    resources = []
    tasks: list[asyncio.Task] = []

    async with asyncio.TaskGroup() as tg:
        if location:
            tasks.append(
                tg.create_task(
                    explore_addresses(project, location, creds),
                    name="explore_addresses",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_gke(project, location, creds), name="explore_gke"
                )
            )
            tasks.append(
                tg.create_task(
                    explore_cloudrun(project, location, creds),
                    name="explore_cloudrun",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_lb(project, location, creds), name="explore_lb"
                )
            )
            tasks.append(
                tg.create_task(
                    explore_connector(project, location, creds),
                    name="explore_connector",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_securities(project, location, creds),
                    name="explore_securities",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_firewalls(project, location, creds),
                    name="explore_firewalls",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_forwardingrules(project, location, creds),
                    name="explore_forwardingrules",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_health_checks(project, location, creds),
                    name="explore_health_checks",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_target_proxies(project, location, creds),
                    name="explore_target_proxies",
                )
            )
        else:
            tasks.append(
                tg.create_task(
                    explore_addresses(project, None, creds),
                    name="explore_addresses",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_sql(project, creds), name="explore_global_sql"
                )
            )
            tasks.append(
                tg.create_task(
                    explore_lb(project, None, creds), name="explore_global_lb"
                )
            )
            tasks.append(
                tg.create_task(
                    explore_vpc(project, creds), name="explore_global_vpc"
                )
            )
            tasks.append(
                tg.create_task(
                    explore_monitoring(project, creds),
                    name="explore_global_monitoring",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_securities(project, None, creds),
                    name="explore_global_securities",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_firewalls(project, None, creds),
                    name="explore_firewalls",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_health_checks(project, None, creds),
                    name="explore_health_checks",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_target_proxies(project, None, creds),
                    name="explore_target_proxies",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_storage(project, creds),
                    name="explore_storage",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_compute(project, creds),
                    name="explore_compute",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_memorystore(project, creds),
                    name="explore_memorystore",
                )
            )
            tasks.append(
                tg.create_task(
                    explore_dns(project, creds),
                    name="explore_dns",
                )
            )

        [t.add_done_callback(task_done) for t in tasks]

    for task in tasks:
        result = task.result()
        if result is None:
            continue
        resources.extend(result)

    name = f"{project}-{location}"

    return Discovery(
        id=make_id(name),
        resources=resources,
        meta=Meta(name=name, display=name, kind="gcp", category=None),
    )


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    cloudrun_links(d, serialized)
    vpc_expand_links(d, serialized)
    sql_expand_links(d, serialized)
    gke_expand_links(d, serialized)
    lb_expand_links(d, serialized)
    mon_expand_links(d, serialized)


###############################################################################
# Private functions
###############################################################################
def task_done(task: asyncio.Task) -> None:
    task.remove_done_callback(task_done)

    if task.cancelled():
        logger.warning(f"Task '{task.get_name()}' cancelled")
        return None

    x = task.exception()
    if x:
        logger.error(f"{x=}")
        return None



# mypy: disable-error-code="call-arg,index"
import logging

import msgspec
from kubernetes import client

from lueur.make_id import make_id
from lueur.models import K8SMeta, Resource
from lueur.platform.k8s.client import AsyncClient, Client

__all__ = ["explore_service"]
logger = logging.getLogger("lueur.lib")


async def explore_service() -> list[Resource]:
    resources = []

    async with Client(client.CoreV1Api) as c:
        services = await explore_services(c)
        resources.extend(services)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_services(c: AsyncClient) -> list[Resource]:
    response = await c.execute("list_service_for_all_namespaces")

    services = msgspec.json.decode(response.data)

    if response.status_code == 403:  # type: ignore
        logger.warning(f"K8S API server access failure: {services}")
        return []

    if "items" not in services:
        logger.warning(f"No services found: {services}")
        return []

    results = []
    for service in services["items"]:
        meta = service["metadata"]
        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=meta["name"],
                    display=meta["name"],
                    kind="service",
                    platform="k8s",
                    ns=meta["namespace"],
                    category="network",
                ),
                struct=service,
            )
        )

    return results



# mypy: disable-error-code="func-returns-value"
import logging
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Any

from lueur.make_id import make_id
from lueur.models import Discovery, Meta
from lueur.platform.aws.ec2 import explore_ec2
from lueur.platform.aws.ecr import explore_ecr
from lueur.platform.aws.eks import explore_eks
from lueur.platform.aws.lb import expand_links as lb_expand_links
from lueur.platform.aws.lb import explore_elbv2
from lueur.platform.aws.vpc import expand_links as vpc_expand_links
from lueur.platform.aws.vpc import explore_vpc

__all__ = ["explore", "expand_links"]
logger = logging.getLogger("lueur.lib")


def explore(region: str | None = None) -> Discovery:
    resources = []
    futures: list[Future] = []

    with ThreadPoolExecutor(max_workers=5) as executor:
        if region:
            futures.append(executor.submit(explore_eks, region))
            futures.append(executor.submit(explore_ec2, region))
            futures.append(executor.submit(explore_ecr, region))
            futures.append(executor.submit(explore_elbv2, region))
            futures.append(executor.submit(explore_vpc, region))
        else:
            pass

        [f.add_done_callback(task_done) for f in futures]

    for future in futures:
        if future.exception():
            continue

        result = future.result()
        if result is None:
            continue

        resources.extend(result)

    name = f"aws-{region}"

    return Discovery(
        id=make_id(name),
        resources=resources,
        meta=Meta(name=name, display=name, kind="aws", category=None),
    )


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    vpc_expand_links(d, serialized)
    lb_expand_links(d, serialized)


###############################################################################
# Private functions
###############################################################################
def task_done(future: Future) -> None:
    if future.cancelled():
        logger.warning("Future cancelled")
        return

    x = future.exception()
    if x:
        logger.error(f"{x=}")



# mypy: disable-error-code="union-attr"
import logging

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import GCPMeta, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client

__all__ = ["explore_firewalls"]
logger = logging.getLogger("lueur.lib")


async def explore_firewalls(
    project: str, location: str | None = None, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://compute.googleapis.com", creds) as c:
        firewalls = await explore_global_firewalls(c, project)
        resources.extend(firewalls)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_global_firewalls(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/compute/v1/projects/{project}/global/firewalls")

    firewalls = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Firewall API access failure: {firewalls}")
        return []

    if "items" not in firewalls:
        logger.warning(f"No global firewalls found: {firewalls}")
        return []

    results = []
    for firewall in firewalls.get("items", []):
        self_link = firewall["selfLink"]
        name = firewall["name"]
        display = name

        results.append(
            Resource(
                id=make_id(firewall["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="global-firewalls",
                    project=project,
                    self_link=self_link,
                    platform="gcp",
                    category="security",
                ),
                struct=firewall,
            )
        )

    return results
